"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from pydelorean.base.node import *
from bs4 import BeautifulSoup
import re
import frontmatter

def get_progressive_expansion(node:Node, **kwargs) -> list:
    
    # BASE CASE: TextNode
    if isinstance(node, TextNode):
        return [(node.corpus, node.corpus)]
    
    if kwargs.get("append", False):
        returned_list = [(node.header, node.corpus)]
    else:
        returned_list = []
    for child in node.children:
        if isinstance(child, HeaderNode):
            child_list = get_progressive_expansion(child, append=True)
            returned_list += [(child.header, para) for header, para in child_list]
        elif isinstance(child, TextNode):
            returned_list.append((node.header, child.corpus))
        
    return returned_list


def find_backlinks(input_text:str) -> list:
    """ 
    Function to find all backlinks in a given text.
    """
    
    backlinks = []

    # Regular expression pattern to match backlinks
    pattern = r'\[\[(.*?)\]\]'
    backlinks = re.findall(pattern, input_text)
    return backlinks


def find_tags(input_text:str) -> list:
    """
    Function to find all tags in a given text.
    """
    
    tags = []

    # Regular expression pattern to match backlinks
    pattern = r'#([a-zA-Z0-9_]+)'
    tags = re.findall(pattern, input_text)
    return tags


def find_metadata(input_text:str):
    post = frontmatter.loads(input_text)
    return post.metadata, post.content


def clean_markdown(md:str, **kwargs) -> str:
        
    remove_header_tags = kwargs.get("remove_header_tags", False)    
    
    # Remove the metadata
    _, markdown_text = find_metadata(md)
    
     # Create a BeautifulSoup object with the input markdown text
    soup = BeautifulSoup(markdown_text, 'html.parser')

    # Remove HTML tags from the document
    markdown_text = soup.get_text()
    
    # Remove links, images, and backlinks
    markdown_text = re.sub(r'!\[(.*?)\]\((.*?)\)', '', markdown_text)
    markdown_text = re.sub(r'\[(.*?)\]\((.*?)\)', r'\1', markdown_text)
    markdown_text = re.sub(r'\[\[(.*?)\]\]', r'\1', markdown_text)
    
    # Remove quotation marks
    markdown_text = markdown_text.replace('>', '')  
    
    # Remove bold and italics formatting
    markdown_text = re.sub(r'\*\*(.*?)\*\*', r'\1', markdown_text)
    markdown_text = re.sub(r'__(.*?)__', r'\1', markdown_text)
    markdown_text = re.sub(r'\*(.*?)\*', r'\1', markdown_text)
    markdown_text = re.sub(r'_(.*?)_', r'\1', markdown_text)
    
    # Remove empty lines
    markdown_text = re.sub(r'^\s*$', '', markdown_text, flags=re.MULTILINE)
    
    # Remove header tags
    if remove_header_tags:
        markdown_text = re.sub(r'#+', '', markdown_text)
    
    return markdown_text




"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from bigtree import Node

# TODO: Add all the documentation and convert to snake case soon.


class TextNode(Node):
    
    FRACTION_PRINTABLE = 0.25   
    
    def __init__(self, name:str, text:str, **kwargs):
        self.text = text
        self.corpus = text
        super().__init__(name, **kwargs)
        
        
    def __getitem__(self, index:int):
        if index < 0 or index >= len(self.text):
            raise IndexError("Index out of bounds")
        return self.text[index]
           
    def __str__(self):
        return self.text[0:min(10, int(self.FRACTION_PRINTABLE * len(self.text)))] + " ..."

    def __repr__(self) -> str:
        return "(Text) " + self.text[0:min(10, int(self.FRACTION_PRINTABLE * len(self.text)))] + " ..."
    

class HeaderNode(Node):
    
    def __init__(self, header:str, headerNumber:int, **kwargs):
        self.header = header
        self.header_level = headerNumber
        self.corpus = header + "\n"
        super().__init__(header, **kwargs)
        
    def __str__(self):
        return self.header
    
    def __repr__(self):
        return f"(h{self._header_level}) {self.header}"
    
    def __and__(self, other):
        self.corpus += "\n" + other.corpus
           

        



"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

class File:
    pass

# FIGUREOUT: Do we really need a file class for each of the supported formats ?
# FIGUREOUT: What more attributes / methods should these classes have ?

class MarkdownFile(File):
    pass

class RestructuredFile(File):
    pass

class AsciidocFile(File):
    pass

class TextFile(File):
    pass

class YAMLFile(File):
    pass

class JSONFile(File):
    pass


FILE_EXTENSIONS = {
    ".md":      MarkdownFile,
    ".rst":     RestructuredFile,
    ".adoc":    AsciidocFile,
    ".txt":     TextFile,
    ".yaml":    YAMLFile,
    ".yml":     YAMLFile,
    ".json":    JSONFile
}


"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from .parser import *


PARSER_CLASSES = {
    ".md"  : MarkdownParser,
    ".rst" : RestructuredParser,
    ".adoc": AsciiDocParser,
    ".yaml": YAMLParser,
    ".yml" : YAMLParser,
    ".json": JSONParser,
    ".xml" : XMLParser
}

def get_parser_from_extension(extension:str) -> Parser:
    """
    Returns the parser class that corresponds to the given extension.
    """
    return PARSER_CLASSES[extension]


def get_parser_from_document_name(document_name:str) -> Parser:
    """
    Returns the parser class that corresponds to the given document name.
    """
    file_extension = document_name.split('.')[-1]
    return get_parser_from_extension(file_extension)


"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from bs4 import BeautifulSoup
from .delorean import *
from .base import Forest
import re


# TODO: Check the type of the input and call the appropriate function
# Use the newly defined File classes for this. 
def treeify(name:str, text:str, *args, **kwargs) -> Forest:
    
    return mdtreeify(name, text, *args, **kwargs)


def textify(forest:Forest, *args, **kwargs) -> str:
    
    return mdtextify(forest)




"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from .files import *
from .base import *
from .parser import *
from .tools import *

# ==================================================================================================
#                                           TREEIFY
# ==================================================================================================

# TODO: Rename to snake_case in the future.

def mdtreeify(name:str, md:str, *args, **kwargs) -> MarkdownForest:
    """
    Converts markdown file to a MarkdownForest
    """
    
    meta, cont = find_metadata(md)
    
    backlinks = find_backlinks(cont)
    tags = find_tags(cont)
    
    if cont == "":
        # If the file is empty, just create a root node
        tree = Node("root")
        return tree
    
    
    parser = MarkdownParser(name, cont)
    tree = parser.parse()
    
    returnForest = MarkdownForest(tree, name, metadata=meta)
    
    for backlink in backlinks:
        returnForest.add_backlink(backlink)
    for tag in tags:
        returnForest.add_tag(tag)

    return returnForest


def rsttreeify(name:str, rst:str, *args, **kwargs) -> RestructuredForest:
    pass


def asciidoc_treeify(name:str, adoc:str, *args, **kwargs) -> AsciiDocForest:
    pass


def yaml_treeify(name:str, yaml:str, *args, **kwargs) -> YAMLForest:
    pass


def jsontreeify(name:str, json:str, *args, **kwargs) -> JSONForest:
    pass


# ==================================================================================================
#                                           TEXTIFY
# ==================================================================================================


def header_node_to_text(rootNode: HeaderNode) -> str:
    
    # BASE CASE: We just have text node.
    if isinstance(rootNode, TextNode):
        return rootNode.text + "\n\n"
    
    # For the header node
    headerLevel = rootNode.header_level
    returnString = str(rootNode) + "\n"
    for _, child in enumerate(rootNode.children):
        returnString += header_node_to_text(child)
    
    return returnString    
     

def convert_to_metadata(metadata_dict:dict) -> str:
    yaml = "---\n"
    for key, value in metadata_dict.items():
        
        if isinstance(value, list):
            yaml += f"{key}: \n"
            for elem in value:
                yaml += f"  - {elem}\n"
            continue
        
        yaml += f"{key}: {value}\n"
    yaml += "---\n"
    return yaml


def mdtextify(forest: MarkdownForest) -> str:
    """
    Converts a MarkdownForest to a markdown file.
    """
    
    returnString = ""
    
    # Get the root node
    root = forest.root
    
    # Get the metadata
    metadata = forest.get_metadata()
    returnString = convert_to_metadata(metadata) + "\n"
    
    for node in root.children:
        returnString += header_node_to_text(node) + "\n"
    
    return returnString


def rsttextify(forest: RestructuredForest) -> str:
    
    pass


def asciidoctextify(forest: AsciiDocForest) -> str:
    pass


def yamltextify(forest: YAMLForest) -> str:
    pass


def jsontextify(forest: JSONForest) -> str:
    pass




"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""


from .utils import *

class Parser:
    
    def __init__(self, document_name:str, text:str):
        self.document_name = document_name
        self.text = text
        
    def parse(self) -> Node:
        pass

   
class MarkdownParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        HEADER_PATTERN = r"^(#+\s+)(.*)"
        
        tree = build_tree(self.text, document_name=self.document_name, header_pattern=HEADER_PATTERN)
        build_corpus(tree)
        return tree
        

class RestructuredParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        HEADER_PATTERN = r"^(=+|-+|`+|'+|\^+|\*+|\.+|~+|\++)\s+(.+)\s+\1$"
        
        tree = build_tree(self.text, document_name=self.document_name, header_pattern=HEADER_PATTERN)
        build_corpus(tree)
        return tree
   

class AsciiDocParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass     


class TextParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass


class YAMLParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass


class JSONParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass
    

class XMLParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass
    



"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from .node import HeaderNode


class Forest:
    
    def __init__(self, root:HeaderNode, document_name:str="[document]"):
        self.document_name = document_name
        self.root = root
        self.num_trees = len(root.children)
        # TODO: Implement a method to count the number of nodes in the tree
    
    def __str__(self):
        return str(self.root)
    
    def __len__(self):
        return self.num_trees
    
    @property
    def name(self):
        return self.document_name
    
    @name.setter 
    def name(self, document_name:str):
        self.document_name = document_name    
    
    def set_root(self, root:HeaderNode):
        self.root = root
    
    # TODO: Implement __str__ and __len__ methods
    # FIGUREOUT: What other methods should this base class have ?


class MarkdownForest(Forest):
    
    # TODO: Make this a subclass of Forest
    def __init__(self, root:HeaderNode, document_name:str="[document]", metadata:dict=None):
        super().__init__(root, document_name)
        self.metadata = metadata
        self.backlinks = []
        self.tags = []

    def get_metadata(self):
        return self.metadata
    
    def add_backlink(self, backlink:str):
        self.backlinks.append(backlink)
    
    def get_backlinks(self):
        return self.backlinks
    
    def add_tag(self, tag:str):
        self.tags.append(tag)
        
    def get_tags(self):
        return self.tags
    

class RestructuredForest(Forest):
    pass


class AsciiDocForest(Forest):
    
    def __init__(self, root:HeaderNode, document_name:str="[document]", document_type:str="article"):
        super().__init__(root, document_name)
        self.document_type = document_type
    
    def get_document_type(self):
        return self.document_type


class TextForest(Forest):
    pass    

class YAMLForest(Forest):
    pass


class JSONForest:
    pass

class XMLForest(Forest):
    pass


"""
    Copyright 2023 Keane Moraes

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import re
from pydelorean.base import TextNode, HeaderNode
from bigtree import Node, print_tree


def parse_text(input_text:str) -> list[TextNode]:
    return_list = []
    
    # Split the text by newlines
    input_text = input_text.strip()
    paragraphs = input_text.split("\n\n")    
    
    for paragraph in paragraphs:
        print(paragraph)
        return_list.append(TextNode(name=paragraph, text=paragraph))
        
    return return_list


def build_tree(inputText:str, *args, **kwargs) -> Node:

    assert isinstance(inputText, str), "inputText must be a string"
    if inputText == "":
        return None
    
    root = HeaderNode(header=kwargs["document_name"], headerNumber=0)

    # Default to the GFM header pattern
    HEADER_PATTERN = kwargs['header_pattern'] if 'header_pattern' in kwargs else r"^(#+\s+)(.*)"
    
    inputText = inputText.strip()
    matches = re.finditer(HEADER_PATTERN, inputText, re.MULTILINE)
        
    # Get all the indices of the headers
    indices = [(match.start(), match.end()) for match in matches]
    
    # BASE CASE
    # Check if it is only text (since there are no header indices)
    if len(indices) == 0:
        list_of_text_nodes = parse_text(inputText)
        for text_node in list_of_text_nodes:
            root >> text_node
            # root ** text_node
        return root
    
    # Is there text before the first header?
    if indices[0][0] != 0:
        textBefore = inputText[0:indices[0][0]]
        list_of_text_nodes = parse_text(textBefore)
        for text_node in list_of_text_nodes:
            root >> text_node
            # root ** text_node

    # Use a stack to keep track of the header levels.
    stack = []

    # Iterate over all the headers and accordingly build the tree.
    for currHeader in indices:

        currHeaderText = inputText[currHeader[0]:currHeader[1]]
        currHeaderLevel = len(currHeaderText) - len(currHeaderText.replace("#", ""))
        
        # If this is the first header then create the header node automatically.
        if len(stack) == 0:
            
            # Create the header node with the current header.
            node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
            root >> node
            # root ** node
                        
            # Add the node to the stack
            stack.append((currHeaderLevel, currHeader, node))
            continue
        else:
            lastHeader = stack[-1]
        
        # CASE 1: If the current header level is greater than the starting header level then keep going.
        if currHeaderLevel > lastHeader[0]:
            
            # Extract the text between the current header and the previous header
            textBetween = inputText[lastHeader[1][1]:currHeader[0]].strip()
            if textBetween != "":
                list_of_text_nodes = parse_text(textBetween)
                for text_node in list_of_text_nodes:
                    lastHeader[-1] >> text_node
                    # lastHeader[-1] ** text_node
                
            # Create the header node and add it to the stack
            node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
            lastHeader[-1] >> node
            # lastHeader[-1] ** node
            
            stack.append((currHeaderLevel, currHeader, node))
            
        # CASE 2: If the current header level is less than or equal to the starting header level then we need to create a new tree.
        elif currHeaderLevel <= lastHeader[0]:
            
            # Get the text between 
            textBetween = inputText[lastHeader[1][1]:currHeader[0]].strip()
            if textBetween != "":
                list_of_text_nodes = parse_text(textBetween)
                for text_node in list_of_text_nodes:
                    lastHeader[-1] >> text_node
                    # lastHeader[-1] ** text_node
                                            
            # Pop off until you get to the node with a lower level than the current header level
            while len(stack) > 0:
                if stack[-1][0] >= currHeaderLevel:
                    stack.pop()
                else:
                    break
            
            if len(stack) == 0:
                
                node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
                root >> node
                # root ** node
            else:
                node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
                stack[-1][-1] >> node
                # stack[-1][-1] ** node
                
            stack.append((currHeaderLevel, currHeader, node))


    # Check if there is any text after the last header
    if indices[-1][1] != len(inputText):
        textEnd = inputText[indices[-1][1]:].strip()
        list_of_text_nodes = parse_text(textEnd)
        for text_node in list_of_text_nodes:
            stack[-1][-1] >> text_node
            # stack[-1][-1] ** text_node
    return root



def build_corpus(root:Node):
    
    if root.is_leaf:
        return 
    
    for child in root.children:
        build_corpus(child)
        root & child
    